part of 'application.dart';

/// Application root widget
class Application extends StatefulWidget {
  final String name;
  final String initialRoute;
  final Set<ActivityRoute> routes;
  final OnUnknownRoute onUnknownRoute;
  final Storage? localeStorage;
  final Set<LabelInterface> labels;
  final LiveEvent<LabelInterface> onLabel;
  final AppFactory appFactory;

  const Application._({
    Key? key,
    required this.name,
    required this.initialRoute,
    required this.routes,
    required this.onUnknownRoute,
    required this.localeStorage,
    required this.labels,
    required this.onLabel,
    required this.appFactory,
  }) : super(key: key);

  @override
  ApplicationState createState() => ApplicationState();

  static ApplicationState of(BuildContext context) =>
      requireNotNull(context.findRootAncestorStateOfType<ApplicationState>(),
          message: () => 'Application not reachable from current context. Have you called runApplication()?');

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties
      ..add(StringProperty('name', name))
      ..add(StringProperty('initialRoute', initialRoute))
      ..add(IterableProperty<ActivityRoute>('routes', routes))
      ..add(ObjectFlagProperty<OnUnknownRoute>.has('onUnknownRoute', onUnknownRoute))
      ..add(DiagnosticsProperty<Storage?>('localeStorage', localeStorage))
      ..add(IterableProperty<LabelInterface>('labels', labels))
      ..add(ObjectFlagProperty<LiveEvent<LabelInterface>?>.has('onLabel', onLabel))
      ..add(ObjectFlagProperty<AppFactory>.has('appFactory', appFactory));
  }
}

class ApplicationState extends State<Application> {
  GlobalKey _key = GlobalKey();

  final LiveEvent<void> onDestroy = LiveEvent();

  final MutableLiveData<bool> isInputMethodVisible = MutableLiveData(initialValue: false);

  final MutableLiveData<AppLifecycleState> lifecycle =
      MutableLiveData(initialValue: WidgetsBinding.instance!.lifecycleState);

  late String _title;

  String get title => _title;

  set title(String? value) => SystemChrome.setApplicationSwitcherDescription(ApplicationSwitcherDescription(
      label: _title = value ?? widget.name, primaryColor: Theme.of(context).primaryColor.value));

  List<Locale> _supportedLocales = const [];

  List<Locale> get supportedLocales => _supportedLocales;

  late Locale _defaultLocale;

  Locale get defaultLocale => _defaultLocale;

  Locale? _enforcedLocale;

  Locale? get enforcedLocale => _enforcedLocale;

  set enforcedLocale(Locale? value) => Dispatcher.postFrame.post(() => setState(() {
        _enforcedLocale = value;
        widget.localeStorage?.let((storage) => value == null
            ? storage.remove('locale')
            : storage['locale'] = '${value.languageCode}_${value.countryCode}');
      }));

  late _AppBindingObserver _binding;

  late AppRouterDelegate _appRouterDelegate;

  void restart() => Dispatcher.postFrame.post(() {
        onDestroy.send(null);
        _appRouterDelegate = _createRouter();
        setState(() => _key = GlobalKey(debugLabel: 'application'));
      });

  @override
  void initState() {
    super.initState();
    _title = widget.name;
    _binding = _AppBindingObserver(this);
    _appRouterDelegate = _createRouter();
    onDestroy.observeForever((_) {
      sessionStorage.removeWhere((key, value) => key.startsWith('_route_'));
      (const GlobalObjectKey(#application).currentContext as ViewModelProvider).clear();
    });
  }

  @override
  Widget build(BuildContext context) {
    String onGenerateTitle(BuildContext _) => title;
    final labelInterfaceDelegate = LabelInterfaceDelegate(widget.labels, widget.onLabel.send);
    _supportedLocales = widget.labels.map((label) => label.locale).toList(growable: false);
    _defaultLocale = basicLocaleListResolution(WidgetsBinding.instance!.window.locales, _supportedLocales);
    widget.localeStorage?.let((storage) {
      final code = storage['locale']?.split('_') ?? const [];
      if (code.length == 2) _enforcedLocale = Locale(code[0], code[1]);
    });
    final root = widget.appFactory(
        _key,
        _appRouterDelegate.provider,
        _appRouterDelegate.parser,
        _appRouterDelegate,
        onGenerateTitle,
        _enforcedLocale,
        <LocalizationsDelegate>[
          labelInterfaceDelegate,
          GlobalMaterialLocalizations.delegate,
          GlobalWidgetsLocalizations.delegate,
          GlobalCupertinoLocalizations.delegate,
        ],
        _supportedLocales);
    assert(() {
      WidgetsApp _findWidgetsApp(GlobalKey key) {
        final app = key.currentWidget;
        if (app is WidgetsApp) return app;
        key = GlobalObjectKey(requireNotNull(_key.currentState,
            message: () =>
                'appFactory must provide key parameter to one of the following widgets: MaterialApp | CupertinoApp | WidgetsApp'));
        return _findWidgetsApp(key);
      }

      Dispatcher.postFrame.post(() {
        final app = _findWidgetsApp(_key);
        if (app.routeInformationProvider != _appRouterDelegate.provider ||
            app.routeInformationParser != _appRouterDelegate.parser ||
            app.routerDelegate != _appRouterDelegate ||
            app.onGenerateTitle != onGenerateTitle ||
            app.localizationsDelegates?.contains(labelInterfaceDelegate) != true ||
            !app.supportedLocales.toSet().containsAll(_supportedLocales))
          throw StateError('MaterialAppFactory must set all given parameters');
      });
      return true;
    }(), 'WidgetsApp parameters check');
    return ViewModelScope(scope: #application, child: root);
  }

  @override
  void dispose() {
    onDestroy.send(null);
    _binding.dispose();
    super.dispose();
  }

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties
      ..add(DiagnosticsProperty<LiveEvent<void>>('onDestroy', onDestroy))
      ..add(DiagnosticsProperty<MutableLiveData<bool>>('isInputMethodVisible', isInputMethodVisible))
      ..add(DiagnosticsProperty<MutableLiveData<AppLifecycleState>>('lifecycle', lifecycle))
      ..add(StringProperty('title', title))
      ..add(IterableProperty<Locale>('supportedLocales', supportedLocales))
      ..add(DiagnosticsProperty<Locale>('defaultLocale', defaultLocale))
      ..add(DiagnosticsProperty<Locale?>('enforcedLocale', enforcedLocale));
  }

  AppRouterDelegate _createRouter() =>
      AppRouterDelegate(this, sessionStorage, widget.routes, widget.initialRoute, widget.onUnknownRoute);
}
